'''
Antivir Gateway Library for Sagator

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj@salstar.sk>

This program is a email antivir gate for postfix.

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

import os,sys,string,re,time
import pwd,grp,socket,rfc822

from default import *
try:
  from config import *
except ImportError:
  print "ERROR! Config file not found! Exiting now."
  sys.exit(1)
import stats
from avlib import *

debug.set_level(DEBUG_LEVEL)

def ctime():
    return time.time()

def dochroot(user,group,newroot):
    def func_err(fn,f):
        try:
          eval(f)
          return 0
        except OSError,(ec,es):
          debug.echo(0,"WARNING! Can't do %10s: %s" % (fn,es))
        except AttributeError:
          debug.echo(0,"WARNING! Can't do "+f+": Your python don't have this function!")
        return 1
    uid,gid=0,0
    if user:
      try:
        uid=pwd.getpwnam(user)[2]
      except KeyError,es:
        debug.echo(0,"ERROR, getpwnam: ",[[user]],es)
        sys.exit(100)
    if group:
      try:
        gid=grp.getgrnam(group)[2]
      except KeyError,es:
        debug.echo(0,"ERROR, getgrnam: ",[[group]],es)
        sys.exit(100)
    if newroot:
      if func_err('chroot',"os.chroot('"+newroot+"')"):
        safe.ROOT_PATH=newroot
    if gid:
      func_err('setgroups',"os.setgroups(["+str(gid)+"])")
      func_err('setgid',"os.setgid("+str(gid)+")")
    if uid:
      func_err('setuid',"os.setuid("+str(uid)+")")
    if (os.getuid()==0) or (os.getgid()==0
        or (os.geteuid()==0) or (os.getegid()==0)):
      debug.echo(0,"ERROR! I can't setuid and i am running as root!")
      sys.exit(10)

def reload_config():
    global SCANNERS
    try:
      import config
      reload(config)
      SCANNERS=config.SCANNERS
      del config
      sn=""
      for scnr in SCANNERS:
        sn=sn+" "+scnr.name
      debug.echo(0,"Configured scanners:",sn)
    except ImportError,es:
      debug.echo(0,"Error reloading scanner config ... ",es)

class smtps(smtp):
  '''
  SMTP server class
  '''
  def __init__(self,conn=0,addr=['stdin']):
    if conn:
      self.f=conn.makefile('rw',self.bufsize)
      self.write=conn.send
      self.readline=self.f.readline
    else:
      self.write=self.send
      self.readline=sys.stdin.readline
    self.FROM=''
    self.TO=[]
    self.RECV=RECV_SMTP
    self.temp=''
    self.t0=ctime()
    self.srv=smtpc()
    self.write(self.srv.readline())
    mail.addr='NOOP CONNECT FROM: '+addr[0]+':'+str(addr[1])\
      +', '+mail.getnamebyaddr(addr[0])+'\r\n'
  def send(self,s):
    sys.stdout.write(s)
    sys.stdout.flush()
  def send_ok(self,r="Ok"):
    self.write("250 "+r+"\r\n")
  def quit_connect(self):
    debug.echo(5,"SMTPS: Closing connect to server")
    self.srv.close()
    while 1:
      line=self.readline()
      if not line: return 1
      if re.search("^QUIT",line,re.IGNORECASE):
        self.write("221 Bye\r\n")
        self.f.close()
        return 1
      else:
        self.write("502 Error: command not implemented\r\n")
        return 0
  def recv(self):
    global mail
    try:
      line=self.readline()
    except IOError,(ec,es):
      debug.echo(0, "smtps(): ",es)
      return 1
    if not line: return 1
    if self.RECV: # receiving data or header
      if ((self.RECV==RECV_HEADER) & ((line=='\r\n') | (line=='\n'))):
        debug.echo(5,"SMTPS: HEADER DONE")
        mail.df.write(line)
        self.RECV=RECV_BODY
      elif (line=='.\r\n') | (line=='.\n'):
        mail.close()
        if len(mail.data)==len(mail.header):
          debug.echo(0,"SMTPS: Warning: No header?")
          mail.header=''
        debug.echo(3,"SMTPS: BODY DONE, size: ",len(mail.data)," B")
        self.RECV=RECV_SMTP
        v,virname=checkvir(self.FROM,self.TO)
        self.tw=ctime()-self.t0
        debug.echo(2,"STATS: "+str(self.tw)+
          " seconds, ",len(mail.data)," bytes, status: ",virname)
        if (v==S_OK) | (v==S_FORCE_SEND):
          debug.echo(5,"SMTPS: Sending data")
          self.srv.conn.send(mail.header)
          self.srv.conn.send(mail.xheader)
          self.srv.conn.send(mail.data[len(mail.header):])
          debug.echo(1,"SMTPS: OK: 250 Ok")
          self.srv.conn.send(line) # send .
          self.write(self.srv.readline())
          stats.update(len(mail.data),self.tw)
        elif v==S_REJECT:
          debug.echo(1,"SMTPS: REJECT: ","550 Content rejected - VIRUS ",virname)
          self.write("550 Content rejected - VIRUS "+str(virname)+"\r\n")
          stats.update(len(mail.data),self.tw,1)
          return self.quit_connect()
        elif v==S_DROP:
          debug.echo(1,"SMTPS: DROP: 250 mail dropped, VIRUS ",virname)
          self.write("250 mail dropped, VIRUS "+str(virname)+"\r\n")
          stats.update(len(mail.data),self.tw,1)
          return self.quit_connect()
        else: # S_TEMPFAIL
          debug.echo(1,"SMTPS: TEMPFAIL: 451 ",virname)
          self.write("451 "+str(virname)+"\r\n")
          stats.update(0,0,0,1) # update fail statistics
          return self.quit_connect()
      else:
        if self.RECV==RECV_HEADER:
          mail.hf.write(line)
        else:
          mail.df.write(line)
    else:
      mail.comm=mail.comm+line
      mailfrom=re.search('^MAIL *FROM:(.*)$',line,re.IGNORECASE)
      rcptto=re.search('^RCPT *TO:(.*)$',line,re.IGNORECASE)
      if mailfrom:
        self.FROM=mailfrom.group(1)
        debug.echo(2,'SMTPS: '+line.rstrip())
      else:
        if rcptto:
          debug.echo(2,'SMTPS: '+line.rstrip())
        else:
          debug.echo(5,'SMTPS: '+line.rstrip())
      if rcptto:
        self.TO.append(rcptto.group(1))
        try:
          mail.recip.append(rfc822.parseaddr(rcptto.group(1))[1])
        except:
          pass
      elif re.search('^DATA',line,re.IGNORECASE):
        self.RECV=RECV_HEADER
      elif re.search('^QUIT',line,re.IGNORECASE):
        self.srv.conn.send(line)
        self.write(self.srv.readline()) # 221 Bye
        return 1
      #if re.search("^[LE]HLO ",line,re.IGNORECASE): # disabling EHLO
      #  self.write("502 Error: command not implemented\r\n")
      #  return 0
      self.srv.conn.send(line)
      self.write(self.srv.conn.recv(BUFSIZE))
      return 0

def do_scan(scnr,fn=''):
    detected,virlist='',[]
    try:
      if fn:
        scnr.filename=fn
      detected,virlist=scnr.scanstream(mail.data)
      for hdr in scnr.xheader():
        mail.addheader(hdr)
      scnr.destroy()
      return detected,virlist,''
    except socket.error,(ec,es):
      debug.traceback(4,"do_scan: ")
      return detected,virlist,scnr.name+": SocketError: "+es
    except (IOError,OSError):
      debug.traceback(4,"do_scan: ")
      return detected,virlist,scnr.name+": Error: "+debug.traceback_value_str()
    except (ScannerError,ValueError),es:
      debug.traceback(4,"do_scan: ")
      return detected,virlist,scnr.name+": "+str(es)

def set_scanners(newscanners):
    global SCANNERS
    SCANNERS=newscanners

def checkvir(sender,recipients):
    try:
      detected,virlist='',[]
      scanarr,scannames=[],'None'
      for scnr in SCANNERS:
        debug.echo(3,"Running: ",scnr.name)
        detected,virlist,scan_reply=do_scan(scnr)
        if scan_reply:
          debug.echo(0, "CheckVir: ERROR: ",scan_reply)
          return S_TEMPFAIL,scan_reply
        else:
          scanarr=scanarr+[scnr.name]
          scannames=string.join(scanarr,' ')
        if detected:
          debug.echo(3,"Satus: ",[detected,virlist])
          break
      mail.addheader("X-Sagator-Scanner", scannames)
      if detected=='':
        return S_OK,'CLEAN' # mail delivered OK
      else: # VIRUS FOUND
        mail.addheader("X-Sagator-Status", detected)
        rstat=globals.ACTION
        debug.echo(5,"action: ",rstat)
        return rstat,detected
    except socket.error,(ec,es):
      debug.echo(0,"checkvir: ERROR: socket.error: ",[[ec]],es)
      debug.traceback(4,"checkvir: ")
      return S_TEMPFAIL,es
    except IOError,(ec,es):
      debug.echo(0,"checkvir: ERROR: IO: ",es)
      debug.traceback(4,"checkvir: ")
      return S_TEMPFAIL,'IOError'
    except:
      debug.traceback(3,"checkvir: ")
      return S_TEMPFAIL,'AnyError, see sagator logs'
