'''
Antivir library for Sagator

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj@salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
    
'''

import os,sys,re,string,time,random,traceback,signal,resource,struct,socket

BUFSIZE=10240
MASTER_PID=os.getpid()
allowed_chars='abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789._-'

class ScannerError(Exception):
      """Any Error Of Any Scanner"""

#########################################################################
### Debuging support

class debug_class:
  '''
  Debug class. Used for debugging informations.
  '''
  debug_level=2
  def __init__(self):
      self.set_logfile('-')
  def dup(self):
      if self.fd>2:
        os.dup2(debug.fd,1)
        os.dup2(debug.fd,2)
  def set_logfile(self,logfile):
      self.logfile=logfile
      if logfile!='-':
        try:
          self.fd=os.open(logfile,os.O_CREAT|os.O_WRONLY|os.O_APPEND,0640)
        except OSError,(ec,es):
          debug.echo(0,"Error, can't open logfile: ",logfile,": ",es) 
        self.logfd=os.fdopen(self.fd,'a',0)
        #os.chmod(logfile,0640)
      else:
        self.logfd=os.fdopen(os.dup(1),'w',0)
        self.fd=1
      sys.stdout=self.logfd
      sys.stderr=self.logfd
  def reopen(self):
      if (self.logfile!='-') and (self.logfile[0:len(self.chroot)]==self.chroot):
        lfn=self.logfile[len(self.chroot):]
        if lfn[0]!="/":
          lfn="/"+lfn
        try:
          fd=os.open(lfn,os.O_CREAT|os.O_WRONLY|os.O_APPEND,0640)
          self.logfd=os.fdopen(self.fd,'a',0)
          os.dup2(fd,self.fd)
          os.close(fd)
          self.dup()
        except:
          debug.echo(0,"Warning, log is not reopened! (permission denied?)")
      else:
        debug.echo(0,"Error: logrotation not possible, log isn't in chroot")
  def set_level(self,level):
      self.debug_level=level
  def echo(self,level,*s):
      if self.debug_level>=level:
        o=''
        for i in s:
          if type(i)==type([]):
            for x in i:
              o=o+str(x)
          else:
            o=o+str(i)
        p=os.getpid()
        while 1:
          try:
            self.logfd.write("%5s: %s\n" % (p,o))
            break
          except IOError,(ec,es):
            if ec!=4:
              break
  def traceback(self,level,arg):
      self.echo(level,arg,traceback.format_exception(sys.exc_type,
                       sys.exc_value, sys.exc_traceback))
  def traceback_value_str(self):
      return string.join(traceback.format_exception_only(
                             sys.exc_type, sys.exc_value),', ')

debug=debug_class()

#########################################################################
### Useful functions

class safe_class:
  ROOT_PATH='/'
  def fn(self,f):
      if f[0]=='/':
        # f.lstrip('/')
        while f[0]=='/':
          f=f[1:]
        return os.path.join(self.ROOT_PATH,f)
      else:
        return f
  def open(self,name,mode='',buffering=-1):
      if buffering>=0:
        return open(self.fn(name),mode,buffering)
      if mode:
        return open(self.fn(name),mode)
      return open(self.fn(name))
  def osopen(self,filename,flag,mode=0777):
      return os.open(self.fn(filename),flag,mode)

safe=safe_class()

def normalize_filename(sfn):
    '''replace !allowed_chars from filename with _'''
    ofn=''
    for chp in range(len(sfn)):
      if not sfn[chp] in allowed_chars:
        ofn=ofn+'_'
      else:
        ofn=ofn+sfn[chp]
    return ofn

class popen:
  '''
  POpen implementation (with some changes)
  '''
  def __init__(self, cmd, wd='', resource_limits={}):
      self.exitstatus,self.killsignal,self.execerror,self.pid=-1,-1,'',0
      self.pcr,self.pcw = os.pipe()
      self.cpr,self.cpw = os.pipe()
      self.oldsigchld=signal.getsignal(signal.SIGCHLD)
      signal.signal(signal.SIGUSR1,self.sighandler)
      signal.signal(signal.SIGCHLD,self.sighandler)
      self.pid = os.fork()
      if self.pid == 0: # child
        signal.signal(signal.SIGUSR1,signal.SIG_IGN)
        signal.signal(signal.SIGCHLD,signal.SIG_IGN)
        os.close(self.pcw)
        os.close(self.cpr)
        os.dup2(self.pcr,0); os.close(self.pcr)
        os.dup2(self.cpw,1)
        os.dup2(self.cpw,2); os.close(self.cpw)
        try:
          if wd:
            os.chdir(wd)
          for key,value in resource_limits.items():
            resource.setrlimit(key,(value,value))
          debug.echo(4,"POpen: Running: ",str(cmd))
          os.execvp(cmd[0],cmd)
        except os.error,(ec,es):
          debug.echo(3,"POpen: os.error: ",es+str([cmd[0]]))
          os.write(1,es+str([cmd[0]]))
        os._exit(10)
      os.close(self.pcr)
      os.close(self.cpw)
      self.tocmd=os.fdopen(self.pcw, 'w', 0)
      self.fromcmd=os.fdopen(self.cpr, 'r', 0)
  def sighandler(self,sn,stack):
      if sn==signal.SIGUSR1:
        self.execerror=os.read(self.cpr,1024)
      else:
        self.wait()
        debug.echo(4,"POpen SIG=",sn,", pid=",self.pid,
              ", exitstatus=",self.exitstatus,
              ", signal=",self.killsignal)
  def wait(self):
      try:
        pid,es=os.waitpid(self.pid,0)
        self.exitstatus=es/256
        self.killsignal=es%256
      except OSError:
        if self.exitstatus<0:
          self.exitstatus=0 # pid already exited, but no exit status
      signal.signal(signal.SIGCHLD,self.oldsigchld)
      return self.exitstatus
  def readlines(self):
      try:
        return self.fromcmd.readlines()
      except IOError,(ec,es):
        if ec==4: #Interrupted system call
          return []
        else:
          raise
  def readline(self):
      try:
        return self.fromcmd.readline()
      except IOError,(ec,es):
        if ec==4: #Interrupted system call
          return []
        else:
          raise
  def write(self,s):
      if self.exitstatus<0:
        return self.tocmd.write(s)
      else:
        return 0
  def close(self,rw=3):
      if rw&1:
        self.tocmd.flush()
        os.close(self.pcw)
      if rw&2:
        os.close(self.cpr)

def rlistdir(rootpath,path=''):
    '''
    list files and directories recursively
    
    return values: [filelist],size
      [filelist] is an array of file and directory names
      size is added filesizes
    '''
    files,fsize=[],0L
    try:
      listdir=os.listdir(safe.fn(os.path.join(rootpath,path)))
    except:
      return [],0L
    for d in listdir:
      stat=os.lstat(safe.fn(os.path.join(rootpath,path,d)))
      if stat[0] & 0x4000: # is_dir
        rld=rlistdir(rootpath,os.path.join(path,d))
        files=files+rld[0]
        fsize=fsize+rld[1]
      else: # is_file or is_symlink or ...
        fsize=fsize+stat[6]
        files.append(os.path.join(path,d))
    return files,fsize

def randomchars(chars='abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'):
    '''
    Return 10 random characters. Characters are randomly selected
    from chars array.
    '''
    s=''
    r=random.Random()
    for i in range(10):
      s=s+chars[int(r.random()*len(chars))]
    return s

class mktemp:
      '''
      make a temporary file or directory
      Makeing of file:
        mktemp('filename-prefix','.suffix','w',0600).autorm()
      Makeinf od directory:
        mktemp('filename-prefix','.suffix','d',0600).autorm()
      '''
      def __init__(self,prefix,suffix='',flags='w',umode=-1):
          self.flag_autorm=0
          self.flags=flags
          while 1:
            s=randomchars()
            self.name=prefix+s+suffix
            self.name_no_prefix=s+suffix
            self.root_name=safe.fn(self.name)
            debug.echo(7,"mktemp: ",self.root_name)
            try:
              if flags=='d':
                if umode<0:
                  umode=0700
                os.mkdir(self.root_name)
                os.chmod(self.root_name,umode)
                if (os.stat(self.root_name)[0] & 0170000)==040000:
                  return
              else:
                if umode<0:
                  umode=0600
                self.fd=os.open(self.root_name,os.O_WRONLY|os.O_CREAT|os.O_EXCL,umode)
                self.f=os.fdopen(self.fd,'w',BUFSIZE)
                if (os.stat(self.root_name)[0] & 0170000)==0100000:
                  return
            except OSError,(ec,es):
              if not (ec in [17,21]):
                if ec in [2,13]:
                  debug.echo(1,"ERROR: Can't make temp file: ",self.root_name,": ",es)
                raise
      def __del__(self):
          if self.flag_autorm:
            self.rm()
      def rm(self):
          if self.flags=='d':
            os.rmdir(self.name)
          else:
            try:
              self.f.close()
            except:
              pass
            os.unlink(self.root_name)
      def autorm(self):
          self.flag_autorm=1
          return self

def tempcleanup(filename,s=''):
    '''
    Safe cleanup for a file.
    '''
    try:
      os.unlink(safe.fn(filename))
    except OSError,(er,es):
      debug.echo(0,s+": ERROR: OSError: ",es,filename)

def rdel(paths):
    '''
    deletes directory recursively
      paths is an array of directories or files
    '''
    for dirname in paths:
      try:
        if os.lstat(safe.fn(dirname))[0] & 0x4000: # is_dir
          for fname in os.listdir(safe.fn(dirname)):
            rdel([os.path.join(dirname,fname)])
          os.rmdir(safe.fn(dirname))
        else: # is fname or symlink
          os.unlink(safe.fn(dirname))
      except OSError,(ec,es):
        if not ec in [2]:
          debug.echo(3,"rdel: ",debug.traceback_value_str())

def fromhdr():
    '''
    Return a "From local@... DATE" string. String is by RFC.
    '''
    return "From local@local.local"+\
      time.strftime("  %c\r\n",time.localtime(time.time()))

def socket_settimeout(sock,seconds):
    '''
    Set a timeout for a socket. Timeout is in seconds.
    '''
    sock.setsockopt(socket.SOL_SOCKET,socket.SO_RCVTIMEO,
      struct.pack('ll',seconds,0))

#########################################################################
### SMTP and mail classes

S_OK=1
S_TEMPFAIL=2
S_REJECT='MAIL REJECTED'
S_DROP='MAIL DROPPED'
S_FORCE_SEND='SENT, sending forced'

RECV_SMTP=0
RECV_HEADER=1
RECV_BODY=2

from cStringIO import StringIO

class mail_class:
  '''
  EMail structure with some functions.
  '''
  addr='' # Connection address
  comm='' # SMTP communication
  header='' # Email Header
  data='' # Email data
  xheader='' # Extended header
  recip=[] # recipient emails
  def __init__(self):
      self.hf=StringIO()
      self.df=StringIO()
  def close(self):
      self.hf.flush()
      self.df.flush()
      mail.header=self.hf.getvalue()
      self.hf.close()
      mail.data=mail.header+self.df.getvalue()
      self.df.close()
  def normalize_header(self,lines):
      ''' normalize to <80 chars per line '''
      SPACE8 = ' '*8
      MAXLINELEN2=78
      MAXLINELEN=MAXLINELEN2-len(SPACE8)
      sublines_semi=[]
      for line in lines.splitlines():
        if len(line.replace("\t", SPACE8)) <= MAXLINELEN2:
          sublines_semi.append(line.rstrip())
        else:
          # split line by semicolons
          while len(line) > MAXLINELEN:
            i=line.rfind(';', 0, MAXLINELEN)
            if i<0:
              break
            sublines_semi.append(line[:i+1].lstrip())
            line=line[i+1:]
          sublines_semi.append(line.lstrip())
        # now split sublines by spaces
        sublines_space=[]
        for line in sublines_semi:
          if len(line.replace("\t", SPACE8)) <= MAXLINELEN2:
            sublines_space.append(line)
          else:
            while len(line) > MAXLINELEN:
              i=line.rfind(' ', 0, MAXLINELEN)
              if i<0:
                break
              sublines_space.append(line[:i])
              line=line[i+1:]
            sublines_space.append(line)
        # and now join it
        sublines=''
        current=sublines_space.pop(0)
        while sublines_space:
          line=sublines_space.pop(0)
          nline=current+line
          if len(nline.replace("\t", SPACE8)) <= MAXLINELEN2:
            current=nline
          else:
            sublines=sublines+current+"\r\n"
            current="\t"+line
        sublines=sublines+current+"\r\n"
        return sublines
  def addheader(self,desc,value=''):
      if value:
        hdr=desc+': '+value+'\r\n'
      else:
        hdr=desc
      self.xheader=self.xheader+self.normalize_header(hdr)
  def hdrlen(self):
      return len(self.header)
  def getnamebyaddr(self,ip):
      try:
        return socket.gethostbyaddr(ip)[0]
      except:
        return ''

# mail object
mail=mail_class()

class SmtpcError(Exception):
  """SMTP Client Error"""
  ConnectError='ConnectError'
  WrongReturnCode='WrongReturnCode'
  SendmailError='SendmailError'

class smtp:
  '''
  Basic SMTP class. All SMTP clases are descended from this class.
  '''
  f=False
  bufsize=BUFSIZE

class smtpc(smtp):
  '''
  SMTP client class.
  '''
  def __init__(self):
    try:
      self.conn=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
      socket_settimeout(self.conn,120)
      self.conn.connect(self.SMTP_SERVER)
    except socket.error,(ec,es):
      debug.echo(0,"SMTPC: ERROR: connect: ",es)
      raise SmtpcError,(SmtpcError.ConnectError,"451 SMTP connection refused\r\n")
    self.f=self.conn.makefile('rw',self.bufsize)
  def __del__(self):
    self.close()
  def send(self,data,rc='...'):
    if data!='':
      debug.echo(5,"SMTPC>: ",data.rstrip())
      self.conn.send(data+"\r\n")
    ret=''
    while 1:
      r=self.readline()
      ret=ret+r
      if re.search("^[0-9][0-9][0-9] ",r):
        break
    debug.echo(5,"SMTPC<: ",ret.rstrip())
    if not re.search("^"+rc+" ",ret):
      raise SmtpcError,(SmtpcError.WrongReturnCode,ret.rstrip()+"!="+rc)
    return ret
  def readline(self):
    return self.f.readline()
  def close(self):
    try:
      self.f.close()
      self.conn.shutdown(2)
      self.conn.close()
    except:
      pass
  def sendmail(self,sender,recipients,ldata=''):
    ret="451 SMTP connection error\r\n"
    try:
      self.send('','220')
      self.send("HELO localhost",'250')
      self.send("MAIL FROM:"+sender,'250')
      for rec in recipients:
        self.send("RCPT TO:"+rec,'250')
      self.send("DATA",'354')
      self.conn.send(ldata)
      debug.echo(5,"SMTPC>: DATA SENT")
      ret=self.send(".","250")
    except SmtpcError.WrongReturnCode,ret:
      raise SmtpcError,(SmtpcError.SendmailError,ret)
    try:
      self.send("QUIT","221")
      self.close()
    except SmtpcError.WrongReturnCode,ret:
      debug.echo(5,"SMTPC: quit error",ret)
    return ret

#########################################################################
### Scanner templates

class globals:
  ACTION=S_REJECT
  QFNAME=''
  FOUND_BY='NotSet'

class ascanner:
  '''
  Default scanner user for building all other realscanners.
  '''
  name='AScanner'
  is_spamscan=0
  is_multistream=0
  filename='stream'
  xhdr=[]
  scanner=0
  #def __init__(self,arg):
  #    self.arg=arg
  def destroy(self):
      pass
  def scanstream(self,stream):
      raise ScannerError,'Not implemented'
  def scanfile(self,files,dirname=''):
      raise ScannerError,'Not implemented'
  def reinit(self):
      pass
  def xheader(self):
      return self.xhdr
  def get(self,var):
      try:
        return eval("self."+var)
      except AttributeError:
        return "UNKNOWN"

class interscanner(ascanner):
  '''
  Default scanner used for building all other interscanners.
  '''
  name='InterScanner'
  scanners=[]
  #def __init__(self,scanner):
  #    self.scanner=scanner
  def rename(self,scanners):
      n=[]
      for s in scanners:
        n.append(s.name)
      self.name=self.name.replace("()","("+string.join(n,',')+")")
      if self.name=="":
        self.name=string.join(n,',')
  def destroy(self):
      self.scanner.destroy()
  def reinit(self):
      self.scanner.reinit()
  def xheader(self):
      return self.scanner.xheader()
  def get(self,var):
      try:
        r=eval("self."+var)
        if not r:
          r=self.scanner.get(var)
        return r
      except AttributeError:
        return self.scanner.get(var)
