'''
Basic scanners for sagator, version 0.5.1

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj@salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
    
'''

from avlib import *
import re,string,os,time,socket

__all__=['regexp_scan','string_scan','smtp_comm',
         'file2stream','stream2file','f2s','s2f','stream2mbox',
         'l_return',
         'match_all','alternatives','match_any',
         'max_file_size','exec_any',
         'report','quarantine','drop','force_send']

class regexp_scan(ascanner):
  '''
  Primitive regexp pattern scanner.
  
  Usage: regexp_scan({'VirName':['RegExp_Pattern...',...],...} [,size])

  Where: 'VirName' is a string, which identifies defined virus
         'RegExp_Pattern...' is a regexp pattern
         size is a number, which defines, how many bytes may be checked.
           If it is not defined, whole stream is scanned.
    
  There can be more patterns for one virus. All patterns
  must match to assign an stream as virus.
  There can also be more virnames in one dictionary.
  '''
  name='RegExpScanner()'
  def __init__(self,regexps,size=0):
      self.regexps=regexps
      self.size=size
  def scanstream(self,stream):
      for vname,regs in self.regexps.items():
        for reg in regs:
          if self.size>0:
            if re.search(reg,stream[:self.size]):
              if reg==regs[-1]:
                self.xhdr.append("X-Sagator-RegExp: "+vname)
                return vname,['regexp: '+vname+' FOUND, '+reg+'\n']
              else:
                break
          else:
            if re.search(reg,stream):
              if reg==regs[-1]:
                self.xhdr.append("X-Sagator-RegExp: "+vname)
                return vname,['regexp: '+vname+' FOUND, '+reg+'\n']
            else:
              break
      return '',[]

class string_scan(ascanner):
  '''
  Primitive string pattern scanner.
  
  Usage: string_scan({'VirName':['Pattern...',...],...} [,size])

  Where: 'VirName' is a string, which identifies defined virus
         'Pattern...' is a string pattern
         size is a number, which defines, how many bytes may be checked.
           If it is not defined, whole stream is scanned.
    
  There can be more patterns for one virus. All patterns
  must match to assign an stream as virus.
  There can also be more virnames in one dictionary.
  '''
  name='StringScanner()'
  def __init__(self,strings,size=0):
      self.strings=strings
      self.size=size
  def scanstream(self,stream):
      for vname,regs in self.strings.items():
        for reg in regs:
          if self.size>0:
            if string.find(stream[:self.size],reg)>=0:
              if reg==regs[-1]:
                self.xhdr.append("X-Sagator-StringScan: "+vname)
                return vname,['string: '+vname+' FOUND, '+reg+'\n']
            else:
              break
          else:
            if string.find(stream,reg)>=0:
              if reg==regs[-1]:
                self.xhdr.append("X-Sagator-StringScan: "+vname)
                return vname,['string: '+vname+' FOUND, '+reg+'\n']
            else:
              break
      return '',[]

class smtp_comm(ascanner):
  '''
  Primitive regexp pattern scanner for SMTP communication.
  
  Usage: smtp_comm({'VirName':['RegExp_Pattern...',...],...})

  Where: 'VirName' is a string, which identifies defined virus
         'RegExp_Pattern...' is a regexp pattern
    
  For more information about this scanner see documentation for
  regexp_scan realscanner.
  
  This scanner also can be used to apply some scnanner to defined
  email adresses. For example:

      match_all(
        smtp_comm({'anydomain:':['^RCPT TO:.*anydomain.dom']}),
        stream2mbox(libclam())
      )
  
  In this example, libclam() scanner is used only if mail is
  adressed to an user on anydomain.dom.
  '''
  name='smtp_comm()'
  def __init__(self,regexps):
      self.regexps=regexps
  def scanstream(self,stream):
      if scan_only.value:
        return '',[]
      for vname,regs in self.regexps.items():
        for reg in regs:
          if re.search(reg,mail.comm,re.IGNORECASE|re.MULTILINE):
            if reg==regs[-1]:
              self.xhdr.append("X-Sagator-SMTP-comm: "+vname)
              return vname,['smtp_comm: '+vname+' FOUND, '+reg+'\n']
          else:
            break
      return '',[]
  def scanfile(self,files,dir=''):
      return self.scanstream('')

#####################################################################
### convertors

class file2stream(interscanner):
  '''
  file2stream interscanner
  
  Usage: file2stream(scanner())
  
  Where: scanner() is an interscanner or realscanner. It must be a streamscanner.
  
  This scanner converts a filescanner input into streamscanner.
  '''
  name='file2stream()'
  def __init__(self,scanner):
      self.rename([scanner])
      self.scanner=scanner
  def scanfile(self,files,dir=''):
      for fn in files:
        f=safe.open(fn)
        vir,ret=self.scanner.scanstream(f.read())
        f.close()
        if vir:
          return vir,ret
      return '',[]

class f2s(file2stream):
  '''An alias for file2stream interscanner'''
  name='f2s()'

class stream2file(interscanner):
  '''
  stream2file interscanner
  
  Usage: stream2file(scanner())
  
  Where: scanner() is a interscanner or realscanner. It must be a filescanner.
  
  You can use it to save streams into files.
  '''
  name='stream2file()'
  is_multistream=1
  def __init__(self,scanner,addheader=0):
      self.rename([scanner])
      self.scanner=scanner
      self.addheader=addheader
      self.filename=''
      self.mboxdir=''
      self.filenames=[]
      self.counter=1
  def destroy(self):
      if self.mboxdir:
        if debug.debug_level<10:
          debug.echo(4,"s2f: destroy():",self.mboxdir)
          rdel([self.mboxdir])
        else:
          debug.echo(4,"s2f: unfinished destroy:",self.mboxdir)
      self.scanner.destroy()
      self.filenames=[]
      self.mboxdir=''
  def creatembox(self,f,data,addheader=0):
      try:
        if addheader:
          f.write(fromhdr())
        f.write(data)
        f.close()
      except os.error,(ec,es):
        debug.echo(0,'creatembox: ERROR: Open: ',es)
        raise
      except IOError,(ec,es):
        debug.echo(0,'creatembox: ERROR: IO: ',es)
        raise
  def addstream(self,stream):
      if stream:
        if self.mboxdir=='':
          self.mbox=mktemp('/tmp/s2f-','.mbd','d',0700)
          self.mboxdir=self.mbox.name
        if self.filename=='':
          self.filename='part-%06d' % (self.counter)
          self.counter=self.counter+1
        mboxfile=mktemp(self.mboxdir+'/','-'+self.filename,'w',0600)
        self.creatembox(mboxfile.f,stream,self.addheader)
        self.filenames.append(mboxfile.name)
  def scanstream(self,stream):
      vir,ret='',[]
      self.addstream(stream)
      try:
        vir,ret=self.scanner.scanfile(self.filenames,self.mboxdir)
      except:
        raise
      return vir,ret
  def scanmultistream(self,streams):
      vir,ret='',[]
      if type(streams)==type({}):
        for filename,stream in streams.items():
          self.filename=filename
          self.addstream(stream)
      else:
        for stream in streams:
          self.addstream(stream)
      try:
        vir,ret=self.scanner.scanfile(self.filenames,self.mboxdir)
      except:
        raise
      return vir,ret

class s2f(stream2file):
  '''An alias for stream2file interscanner.'''
  name='s2f()'

class stream2mbox(stream2file):
  '''An alias for stream2file(...,1) interscanner.'''
  name='stream2mbox()'
  def __init__(self,scanner):
      stream2file.__init__(self,scanner,1)

#####################################################################
### primitive class

class l_return(ascanner):
  '''
  l_return() realscanner
  
  Usage: l_return('VirName',['virreturnstringline1',...])
      or l_return()
  
  Where: 'VirName',[...] is a returned virus.
  
  This scanner has no real functionality. It return always
  a defined virus, or clean (if virus is not defined).
  This scanner has no error codes.
  '''
  name='l_return()'
  def __init__(self,vir='',ret=[]):
      self.vir=''+vir
      self.ret=[]+ret
      self.filename=''
  def scanstream(self,stream):
      if self.filename:
        debug.echo(3,self.name,": Scanning file: ",self.filename)
      return self.vir,self.ret
  def scanfile(self,files,dirname=''):
      for fn in files:
        debug.echo(3,self.name,": Scanning file: ",fn)
      return self.vir,self.ret

#####################################################################
### match classes

class match_class(interscanner):
  '''parent class for all match clases'''
  def __init__(self,*scanners):
      if type(scanners[0])==type(()):
        self.scanners=scanners[0]
      else:
        self.scanners=scanners
      self.rename(self.scanners)
  def destroy(self):
      for scnr in self.scanners:
        scnr.destroy()
  def reinit(self):
      for scnr in self.scanners:
        scnr.reinit()
  def xheader(self):
      xhdr=[]
      for scnr in self.scanners:
        xhdr.extend(scnr.xheader()) 
      return xhdr
  def get(self,var):
      try:
        r=eval("self."+var)
        if not r:
          raise AttributeError
      except AttributeError:
        for scnr in self.scanners:
          try:
            r+=scnr.get(var)
          except NameError:
            r=scnr.get(var)
          except AttributeError:
            r=''
      return r

class match_all(match_class):
  '''
  match_all() interscanner
  
  Usage: match_all(scanner1(),scanner2(),...)
  
  Where: scanner1(),... are scanners, which must find a virus
  
  Returns a virus only if all scanners returns a virus.
  '''
  name='match_all()'
  def scanstream(self,stream):
      vir,ret='',[]
      for self.scanner in self.scanners:
        v,r=self.scanner.scanstream(stream)
        if v:
          if vir:
            vir=vir+','+v
          else:
            vir=v
          ret=ret+r
        else:
          return '',[]
      return vir,ret

class alternatives(match_class):
  '''
  alternatives() interscanner
  
  Usage: alternatives(scanner1(),scanner2(),...)
  
  Where: scanner1(),... are scanners, which may find a virus
  
  Returns a virus, if one scanner returns a virus.
  Scannig is ended if one scanner returns virus or clean.
  Next scanner from the list is used, if previous scannes fails
  (raises an error). Raises an ScannerError, if all scanners failed.
  '''
  name='alternatives()'
  def scanstream(self,stream):
      vir,ret='',[]
      for self.scanner in self.scanners:
        try:
          vir,ret=self.scanner.scanstream(stream)
          return vir,ret
        except:
          debug.traceback(4,'')
          debug.echo(3,self.name,": scanner ",self.scanner.name," failed: ",
                string.rstrip(debug.traceback_value_str()))
      debug.echo(3,self.name,": All scanners failed")
      raise ScannerError,''

class match_any(match_class):
  '''
  match_any() interscanner
  
  Usage: match_any(scanner1(),scanner2(),...)
  
  Where: scanner1(),... are scanners, which may find a virus
  
  Returns a virus, if one scanner returns a virus.
  Next scanner from list is used, if previous scanner returns
  no virus, or failed (raised an error).
  '''
  name='match_any()'
  def scanstream(self,stream):
      vir,ret='',[]
      for self.scanner in self.scanners:
        try:
          vir,ret=self.scanner.scanstream(stream)
          if vir:
            return vir,ret
        except:
          debug.traceback(4,'')
          debug.echo(3,self.name,": scanner ",self.scanner.name," failed: ",
                string.rstrip(debug.traceback_value_str()))
      return '',[]

#####################################################################
### recover class

class recover(match_any):
  '''
  recover() interscanner
  
  Usage: recover(scanner1(),scanner2(),...)
  
  Where: scanner1(),... are scanners, which may find a virus
  
  Returns a status of first scanner, which returns virus.
  If an error is raised from last scanner, this is cleaned
  and no virus is returned. This can be used, if you want to pass
  any file (virus and spam too) to recipient on error.
  '''
  name='recover()'
  def scanstream(self,stream):
      try:
        return match_any.scanstream(self,stream)
      except:
        debug.echo(2,"ERROR: recover(): All scanners failed!")
        return '',[]

#####################################################################
### max_file_size class

class max_file_size(ascanner):
  '''
  max_file_size() realscanner
  
  Usage: max_file_size(size[,'name'])
  
  Where: size is a number of bytes, which if exceeded, virus is returned
         'name' is a string, which identifies the virus name.
            If it is not present, 'FileSizeOverrun' is returned.
  '''
  name='max_file_size()'
  def __init__(self,size,vname='FileSizeOverrun'):
      self.size=size
      self.vname=vname
  def scanstream(self,stream):
      if len(stream)<=self.size:
        return '',[]
      else:
        return self.vname,[self.filename+": "+self.vname]
  def scanfile(self,files,dir=''):
      for fn in files:
        if os.lstat(safe.fn(fn))[6]>self.size:
          return self.vname,[fn+': '+self.vname]
      return '',[]

#####################################################################
### exec_any class

class exec_any(ascanner):
  '''
  exec_any() realscanner
  
  Usage: exec_any(['command','arg1',...],[statusA1,...],[statusB1,...])
  
  Where: 'command' is a string, which command to run
         'argX' are command line arguments, as last argument the
           filename is added
         statusA1,... are integers, which defines "clean" exit codes
         statusB1,... are integers, which defines "virus found" exit codes
  
  If command returns exit status 0, no virus is returned.
  If exit code is any from statusX, virus is returned.
  Otherwise an error is raised.
  '''
  name='exec_any()'
  def __init__(self,arg,clean_status=[0],infected_status=[1]):
      self.arg=arg
      self.clean_status=clean_status
      self.infected_status=infected_status
  def scanfile(self,files,dir=''):
      for fn in files:
        pf=popen(self.arg+[fn])
        pf.wait()
        pf.close()
        if pf.exitstatus in self.clean_status:
          return '',[]
        else:
          if pf.exitstatus in self.infected_status:
            return 'UnknownVirus',['ExitCode: '+str(pf.exitstatus)]
          else:
            raise ScannerError,'UnkonwnExitStatus: '+str(pf.exitstatus)

#####################################################################
### REPORT class

# MESSAGE REPORT TEMPLATE
REPORT_MSG_TEMPLATE='''\
From: $FROM
To: $TO
Subject: ALERT, $VIRNAME FOUND!

                             VIRUS FOUND!

---------------------------------------------------------------------
SMTP COMMUNICATION:
$SMTP_COMM
---------------------------------------------------------------------
STATUS: $STATUS
QUARANTINED AS: $QNAME
---------------------------------------------------------------------
OUTPUT OF SCANNER:
$SCANNER_OUTPUT
---------------------------------------------------------------------
HEADER OF MESSAGE:
$MSG_HEADER
'''

class scan_only:
  value=False

class report(match_any):
  '''
  report() interscanner - report any message to admin, user or anybody
  
  usage: report(recipients, msg, scanners)
  or:    report(...).sender('email@of.sender.dom')

  where: recipients is an array of recipients, like: ['r1@dom','r2@dom',...]
         msg is an message template, you can use report.MSG_TMPL

  This module can be used to report an message to administrator
  or to virus/spam sender.
  '''
  # Send virus reports to ['...','...']
  RECIPIENTS_TMPL=['root@localhost']
  # Templates
  MSG_TMPL=REPORT_MSG_TEMPLATE
  # scanner
  name='report()'
  def __init__(self,recipients,msg,*scanners):
      self.ADMIN_MSG=msg
      self.USER_MSG=''
      self.RECIPIENTS=recipients
      self.SENDER='vscan@'+socket.gethostname()
      match_any.__init__(self,scanners)
  def sender(self,email):
      self.SENDER=email
      return self
  def replace_tmpl(self,msg,vars):
      idx={}
      for v in vars.keys():
        try:
          idx[msg.index(v)]=v
        except ValueError:
          pass
      keys=idx.keys() # replace it from last to first
      keys.sort()
      keys.reverse()
      for k in keys:
        msg=re.sub("\$"+idx[k],vars[idx[k]],msg,1) # replace only 1st match
      return msg
  def scanstream(self,stream):
      detected,virlist=match_any.scanstream(self,stream)
      if scan_only.value:
        return detected,virlist
      if not detected:
        return detected,virlist
      scanout=''
      for i in virlist:
        scanout+=i
      repl_vars={
        'FROM':		self.SENDER,
        'VIRNAME':	detected,
        'SMTP_COMM':	mail.addr+mail.comm.rstrip(),
        'STATUS':	globals.ACTION,
        'QNAME':	globals.QFNAME,
        'SCANNER_OUTPUT': "SCANNER: "+self.scanner.name+"\n"+scanout,
        'MSG_HEADER':	mail.header.rstrip()
      }
      if self.RECIPIENTS:
        repl_vars['TO']=string.join(self.RECIPIENTS,',\r\n\t')
        msg=self.replace_tmpl(self.ADMIN_MSG,repl_vars)
        try:
          smtpc().sendmail(self.SENDER,self.RECIPIENTS,msg)
        except SmtpcError,(ses,es):
          debug(1,"report(): AdminReport: ",ses)
      if self.USER_MSG:
        repl_vars['TO']=string.join(mail.recip,',\r\n\t')
        msg=self.replace_tmpl(self.USER_MSG,repl_vars)
        try:
          smtpc().sendmail(self.SENDER,self.RECIPIENTS,msg)
        except SmtpcError,(ses,es):
          debug(1,"report(): UserReport: ",ses)
      return detected,virlist

#####################################################################
### QUARANTINE class

class quarantine(match_any):
  '''
  quarantine() - quarantine into any directory
  
  usage: quarantine('qdir','dont quarantine regular expression',scanners)

  This scanner can be used to quarantine an virus/spam.
  qdir is a directory, in which viruses are stored.
  Following regular expression can be used to disable quarantining
  of some viruses or spams.
  '''
  name='quarantine()'
  def __init__(self,qdir,dontq,*scanners):
      self.QDIR=qdir
      self.DONTQ=dontq
      match_any.__init__(self,scanners)
  def scanstream(self,stream):
      detected,virlist=match_any.scanstream(self,stream)
      if scan_only.value:
        return detected,virlist
      if detected:
        if (self.QDIR!='') & (self.DONTQ!='') \
            & (re.search(self.DONTQ,detected,re.IGNORECASE)!=None):
          qfname='Not quarantined yet.'
        else:
          fext=time.strftime("-%Y%m%d-%H%M%S-",time.localtime(time.time()))
          fpath=self.QDIR+'/qm'
          num=0
          while 1:
            qfname=fpath+fext+str(num)
            num+=1
            try:
              fd=safe.osopen(qfname,os.O_WRONLY|os.O_CREAT|os.O_EXCL,0660)
              f=os.fdopen(fd,'w',BUFSIZE)
              f.write(mail.addr)
              f.write(mail.comm)
              f.write(mail.xheader)
              f.write(mail.data)
              f.write(".\r\nQUIT\r\n")
              f.close()
              break
            except OSError,(ec,es):
              if ec==17: # already exists
                continue
              else:
                debug.echo(2,"ERROR: quarantine(): ",es)
                return detected,virlist
        globals.QFNAME=qfname
      return detected,virlist

#####################################################################
### drop and force_send classes

class drop(match_any):
  '''
  drop() - a class to DROP viruses/spams, by default they are REJECTED

  usage: drop(drop_pattern,scanners)
  
  This scanner can be used to drop some viruses, like Klez.
  Klez sends fake sender address and this virus can't be sent
  back to sender, because sender is faked.
  '''
  name='drop()'
  def __init__(self,drop_pattern,*scanners):
      self.DROP=drop_pattern
      match_any.__init__(self,scanners)
  def scanstream(self,stream):
      detected,virlist=match_any.scanstream(self,stream)
      if re.search(self.DROP,detected,re.IGNORECASE):
        globals.ACTION=S_DROP
      return detected,virlist

class force_send(match_any):
  '''
  force_send() - a class to force sending some viruses/spams

  usage: force_send(force_send_pattern,scanners)
  
  This scanner can be used to send some viruses/spams.
  You can define by regular expression, what virus name or spam
  can be sent.
  '''
  name='force_send()'
  def __init__(self,force_send_pattern,*scanners):
      self.FORCE_SEND=force_send_pattern
      match_any.__init__(self,scanners)
  def scanstream(self,stream):
      detected,virlist=match_any.scanstream(self,stream)
      if (self.FORCE_SEND!='') \
        & (re.search(self.FORCE_SEND,detected,re.IGNORECASE)!=None):
          globals.ACTION=S_FORCE_SEND
      return detected,virlist
