'''
nodold module, version 0.3.0

(c) 2003 Jan ONDREJ (SAL) <ondrejj@salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from avlib import *
import os,socket,re,string

class nodold(ascanner):
  '''
  NOD32d scanner. This works with nod32d from ESET.
  
  Usage: nodoldd(['host',port],params)
  
  Where: 'host' is a string, which defines a host of nod32d
            (by default 'localhost')
         port is a number, which defines a port of nod32d
            (by default 8448)
         params is a string to define scanner parameters.
            It's defaults to "@patt:1@heur:2@clean:0".
            There is no format description.
  
  This scanner mat be used to scan for viruses by ESET's NOD32.
  Only old version (buyed directly from ESET) may be used.
  
  WARNING! NOD32 can't scan direct mboxes and archives.
  You need mimeparse() module to parse mailboxes. For exmple:
    mimeparse(stream2file(nodoldd(['localhost',8448])))
  If you need also unpacker support, add decompress interscanner:
    mimeparse(stream2file(decompress(nodoldd(['localhost',8448]))))
  
  If you want start nod32d in chroot, use for example:
    exec chroot /var/spool/vscan /opt/nod32/nod32d
  from any script.
  '''
  name='NodOldD()'
  def __init__(self,arg,params="@patt:1@heur:2@clean:0"):
      self.arg=arg
      self.params=params
  def scanfile(self,filenames,dirname=''):
      debug.echo(7,"nodold(): ",[filenames],dirname)
      if type(self.arg)==type('string'): # local socket
        s=socket.socket(socket.AF_UNIX,socket.SOCK_STREAM)
        socket_settimeout(s,120)
        s.connect(self.arg)
      else:
        s=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
        socket_settimeout(s,120)
        s.connect((self.arg[0],self.arg[1]))
      f=s.makefile('rw',BUFSIZE)
      r=f.readline() # 200 NOD32SS 1.386 (20030326)
      s.send("PARM "+self.params+"\n")
      r=f.readline() # 200 OK (patt: 1, heur: 2, clean: 0)
      ret=[]
      for fname in filenames:
        s.send("SCAN "+os.path.join(dirname,fname)+"\n")
        line=f.readline()
        if not re.search("^200",line):
          ret.append(line[4:])
        if re.search("^5",line): # error opening file?
          raise ScannerError,'nodold: '+line
      s.send("QUIT\n")
      r=f.readline()
      f.close()
      s.shutdown(2)
      s.close()
      if ret==[]:
        vir=''
      else:
        vir=string.rstrip(ret[0])
        self.xhdr.append("X-Sagator-nod32: "+vir)
      return vir,ret
