#!/usr/bin/python2

'''
sagator.py
(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj@salstar.sk>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from aglib import *
import signal
import avlib,stats

TEST_MSG='''\
From: test@tester.test\r
To: test2@tester.test\r
Subject: testing mail\r

simple mail\r
'''

stat_coll=0

def sigchld(sn,stack):
    pid,s=os.wait()
    debug.echo(3,"Child exited ... pid,status: ",(pid,s/256,s%256))

def sighup(sn,stack):
    debug.echo(3,"SIGHUP: Reloading configuration ...")
    reload_config()
    debug.echo(3,"SIGHUP: Reinitializing scanners ...")
    for scanner in SCANNERS:
      scanner.reinit()

def sigusr1(sn,stack):
    debug.echo(3,"SIGUSR1: Reopening log ...")
    debug.reopen()

def sigterm(sn,stack):
    try:
      os.unlink(PID_FILE)
    except:
      pass
    if stat_coll.masterpid>0:
      try:
        os.kill(stat_coll.masterpid,signal.SIGTERM)
      except:
        pass
    debug.echo(1,"SERVER: Exiting - SIGTERM ...")
    sys.exit(0)

def test_scanners():
    scanarr,scannames=[],'None'
    mail.data=TEST_MSG
    for scnr in SCANNERS:
      if scnr.name:
        debug.echo(1, "Testing "+scnr.name+"...")
      detected,virlist,scan_report=do_scan(scnr)
      if scan_report:
        debug.echo(0, "Scanner ",scnr.name," test failed! Disable it manually!")
        debug.echo(0, "  ",scan_report)
      scanarr=scanarr+[scnr.name]
    mail.data=''
    #debug.echo(4,"Configured scanners: ",string.join(scanarr,' '))

def main(host,port,daemon=0):
    global stat_coll
    debug.echo(1,"SERVER: Starting ...")
    try:
      if port<0:
        dochroot(USER,GROUP,CHROOT)
        m=smtps()
        while 1:
          if m.recv(): break
        sys.exit(0)
      else:
        s=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET,socket.SO_REUSEADDR,1)
        try:
          s.bind((host,port))
          s.listen(2)
        except socket.gaierror,(ec,es):
          debug.echo(0,"SERVER: ERROR: BIND ",es)
          sys.exit(1)
        # initialize statistics collector (bind to its address)
        if COLLECTOR_SERVER:
          stat_coll=stats.collector(COLLECTOR_SERVER)
        try:
          pidf=open(PID_FILE,'w')
        except:
          debug.echo(0,"Can't open pid file: ",PID_FILE)
          pidf=0
        dochroot(USER,GROUP,CHROOT)
        # reinit scanners (load virdb...)
        for scnr in SCANNERS:
          scnr.reinit()
        # do test as another process
        if os.fork()==0:
          test_scanners()
          sys.exit(0)
        os.wait()
        # and now start the daemon
        signal.signal(signal.SIGCHLD,sigchld)
        signal.signal(signal.SIGHUP,sighup)
        signal.signal(signal.SIGUSR1,sigusr1)
        signal.signal(signal.SIGTERM,sigterm)
        debug.echo(1,"sagator started, waiting for connections ...")
        if daemon:
          os.close(0)
          debug.dup()
          if os.fork()>0:
            debug.echo(1,"daemonizing...")
            sys.exit(0)
        if pidf:
          pidf.write(str(os.getpid()))
          pidf.close()
        # start statistics collector
        if COLLECTOR_SERVER:
          stat_coll.main()
        while 1:
          try:
            conn,addr = s.accept()
            avlib.socket_settimeout(conn,120)
            pid=os.fork()
            if pid==0:
              signal.signal(signal.SIGCHLD,signal.SIG_IGN)
              signal.signal(signal.SIGHUP,signal.SIG_IGN)
              signal.signal(signal.SIGUSR1,signal.SIG_IGN)
              debug.echo(1,"SERVER: Connection from: "+addr[0]+" at "+\
                time.strftime("%c",time.localtime()))
              m=smtps(conn,addr)
              while 1:
                if m.recv(): break
              conn.shutdown(2)
              debug.echo(1,"SERVER: Closing connection.")
              conn.close()
              sys.exit(0)
            conn.close()
          except SmtpcError,(ses,es):
            # error is written, only exit now
            sys.exit(2)
          except socket.error,(en,es):
            if en==107: # 'Transport endpoint is not connected'
              debug.echo(1,"SERVER: Connection closed by peer.")
              sys.exit(0)
            elif en==32: # 'Broken pipe'
              debug.echo(0,"SERVER: ERROR: Broken pipe.")
              sys.exit(1)
            elif en!=4:
              debug.echo(0,'SERVER: ERROR: SocketError: ',es)
              debug.traceback(0,"SERVER: ")
              if en==9:
                sys.exit(1)
    except KeyboardInterrupt:
      try:
        os.unlink(PID_FILE)
      except:
        pass
      debug.echo(1,"SERVER: Exiting - KeyboardInterrupt ...")
      sys.exit(0)

if __name__ == '__main__':
  host=BIND_ADDR[0]
  port=BIND_ADDR[1]
  daemon=0
  n=1
  while n<len(sys.argv):
    if sys.argv[n]=='--help':
      print "SAGATOR 0.3"
      print "Usage: --help        this help"
      print "       --reload      reinitialize scanners"
      print "       --inetd       stard as inetd service (useful for postfix too)"
      print "       --daemon      daemonize after start"
      print "       --host h      smtp server host"
      print "       --port p      smtp server port"
      print "       --logfile l   filename for logging"
      print "       --debug l     set debug level to l"
      sys.exit(0)
    elif sys.argv[n]=='--reload':
      try:
        pid=int(open(PID_FILE,'rt').readline())
      except:
        print "Error: Can't read pid!"
        sys.exit(1)
      try:
        os.kill(pid,signal.SIGHUP)
      except:
        print "Error: Unable to send sighup to process %d!" % (pid)
        sys.exit(2)
      sys.exit(0)
    elif sys.argv[n]=='--inetd':
      host='-'
      port=-1
      n=n+1
    elif sys.argv[n]=='--daemon':
      daemon=1
      n=n+1
    elif sys.argv[n]=='--host':
      host=sys.argv[n+1]
      n=n+2
    elif sys.argv[n]=='--port':
      port=int(sys.argv[n+1])
      n=n+2
    elif sys.argv[n]=='--logfile':
      LOGFILE=sys.argv[n+1]
      n=n+2
    elif sys.argv[n]=='--debug':
      debug.set_level(int(sys.argv[n+1]))
      n=n+2
    else:
      print "Unknown parameter:", sys.argv[n]
      sys.exit(1)
  debug.set_logfile(LOGFILE)
  debug.chroot=CHROOT
  avlib.smtp.SMTP_SERVER=SMTP_SERVER
  main(host,port,daemon)
