'''
spamassassin module, version 0.3.4

(c) 2003 Jan ONDREJ (SAL) <ondrejj@salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

from avlib import *
import re,socket

__all__=['spamassassin','spamassassind']

class spamassassin(ascanner):
  '''
  SpamAssassin scanner
  
  Usage: spamassassin(['/usr/bin/spamassassin','-e'],SPAMSTRING,SA_MAX_FILE)
  
  Where: in [] is a path and argumets of spamassassin binary
         SPAMSTRING is a string, which define a spamassassin report,
           which is checked. If it is found, mail is assigned as 'SPAM'
           (by default all reports with exit status!=0 are reported as spam).
         SA_MAX_FILE is a number, which defines maximum file size,
           which can be tested. If a mail (with header) is larger,
           it is not checked (by default 500000).

  This scanner scans for spams. If a spam is returned, a string
  SPAM is returned as virus name. This scanner uses a spamassassin
  binary. For better performance, configure spamd and use
  spamassassind.
  
  WARNING! You need to install spamassassin into chroot manually.
  mkchroot.sh script does it not. If you want to use spamassasin
  with chroot and don't want to make a copy in chroot, use
  a spamassassind scanner, but remember, that scanning for spams
  are not done in chroot environment.
  '''
  is_spamscan=1
  name='SpamAssassin()'
  xhdr=[]
  def __init__(self,arg,spamstring='',sa_max_file=500000):
      self.arg=arg
      self.SPAMSTRING=spamstring
      self.SA_MAX_FILE=sa_max_file
  def scanstream(self,stream):
      exitstatus=''
      if len(stream)>self.SA_MAX_FILE:
        return '',['File too long to test it. Size: '+str(len(stream))]
      pf=popen(self.arg)
      pf.write(stream)
      pf.tocmd.close()
      pf.wait()
      ret,exitstatus,lastspamline=[],'',0
      for line in pf.readlines():
        if re.search("^$",line):
          break
        if re.search("^\t",line):
          if lastspamline:
            ret.append(ret.pop(-1)+line)
        else:
          if re.search("^X-Spam-",line):
            ret.append(line)
            lastspamline=1
            if self.SPAMSTRING:
              if re.search(self.SPAMSTRING,line):
                exitstatus='SPAM'
          else:
            lastspamline=0
      self.xhdr=ret
      pf.fromcmd.close()
      if (pf.exitstatus>0) and (self.SPAMSTRING==''):
        exitstatus='SPAM'
      return exitstatus,ret

class spamassassind(ascanner):
  '''
  SpamAssassinD scanner
  
  Usage: spamassassind(['localhost',783],REQ_SPAM_LEVEL,SA_MAX_FILE)

  Where: ['localhost',783] are the host and port of spamd.
         if REQ_SPAM_LEVEL is less than spam hits, virus is returned,
           if it is -1 (default), spam level from spamassassin config is used
         for SA_MAX_FILE see spamassassin scanner description.
  
  This scanner scans for spams. If a spam is returned, a string
  SPAM is returned as virus name. This scanner uses a spamassassind
  daemon. If you are not familiar with it, you also can use
  spamassassin binary. Look at spamassassin scanner.
  '''
  is_spamscan=1
  name='SpamAssassinD()'
  def __init__(self,arg,reqspamlevel=-1,sa_max_file=500000):
      self.arg=arg
      self.REQSPAMLEVEL=reqspamlevel
      self.SA_MAX_FILE=sa_max_file
      self.xhdr=[]
  def scanstream(self,stream):
      if len(stream)>self.SA_MAX_FILE:
        return '',['File too long to test it. Size: '+str(len(stream))]
      s=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
      socket_settimeout(s,120)
      s.connect((self.arg[0],self.arg[1]))
      f=s.makefile('rw',BUFSIZE)
      s.send("SYMBOLS SPAMC/1.2\r\n")
      s.send("Content-length: "+str(len(stream))+"\r\n")
      s.send("User: vscan\r\n\r\n")
      s.send(stream)
      s.shutdown(1)
      s.close()
      output=f.readlines()
      debug.echo(4,"Spamd output: ",[[len(stream)]],str(output))
      # spamd returned any error?
      if not re.search('^SPAMD/[0-9.]* 0 ',output[0]):
        raise ScannerError,'SpamAssassind: '+output[0]
      # workaround for 8bit mails
      if re.search('^SPAMD/[0-9.]* 76 ',output[0]):
        debug.echo(3,"Spamd WARNING: "+output[0])
        return '',[]
      # spamd returned EX_OK
      reg1=re.search('^Spam: ([a-zA-Z0-9]*) ; ([-0-9.]*) / ([0-9.]*)$',
                     string.rstrip(output[1]))
      if len(output)>3:
        spam_tests=string.rstrip(output[3])
      else:
        spam_tests=""
      if not reg1:
        debug.echo(0,"SPAMASSASSIND: ERROR: ",output)
        raise ScannerError,'SpamAssassind: protocol error'
      spam_yn=reg1.group(1)
      spam_hits=float(reg1.group(2))
      spam_req=float(reg1.group(3))
      spam_level='*'*int(spam_hits)
      if self.REQSPAMLEVEL>=0:
        if spam_hits>=self.REQSPAMLEVEL:
          detected="SPAM_LEVEL:"+spam_level
        else:
          detected=""
      else:
        if spam_yn=="True":
          spam_yn="Yes"
          detected="SPAM_LEVEL:"+spam_level
        if spam_yn=="False":
          spam_yn="No"
          detected=''
      x_spam_status="X-Spam-Status: %s, hits=%f, required=%f\r\n\ttests=%s\r\n"\
        % (spam_yn,spam_hits,spam_req,spam_tests)
      x_spam_level="X-Spam-Level: "+spam_level+"\r\n"
      self.xhdr=[x_spam_status,x_spam_level]
      debug.echo(3,"Spamd status: spam=%s, hits=%f" % (spam_yn,spam_hits))
      return detected,[x_spam_status,x_spam_level]
