'''
Sagator's statistics collector and distributor, version 0.2.2

(c) 2003-2004 Jan ONDREJ (SAL) <ondrejj@salstar.sk>
                                                                                
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

'''

import socket,re,string,os,sys
import urllib,time,signal,struct,atexit
import base
from avlib import *

__all__ = ['collector','update','query','status']

ACCESSPASS='NULL'
COLLECTOR_SERVER=()

def uptime(starttime):
    up=time.mktime(time.localtime())-starttime
    mins=int(up/60)%60
    hours=int(up/3600)%24
    days=int(up/86400)
    if days>0:
      return "%d days, %02d:%02d" % (days,hours,mins)
    else:
      return "%02d:%02d" % (hours,mins)

class collector:
  data={ 'Total-count':0.0, 'Total-bytes':0.0,
         'Filtered-count':0.0, 'Filtered-bytes':0.0,
         'Total-time':0.0 }
  childs=[]
  def __init__(self,bindaddr):
      global ACCESSPASS
      # startup initialization
      self.bindaddr=bindaddr
      ACCESSPASS=randomchars()
      # create a server
      self.s=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
      self.s.setsockopt(socket.SOL_SOCKET,socket.SO_REUSEADDR,1)
      self.s.bind(bindaddr)
      self.s.listen(10)
  def main(self):
      global COLLECTOR_SERVER
      COLLECTOR_SERVER=self.bindaddr
      self.lastsave=time.mktime(time.localtime())
      self.data['Start-at']=self.lastsave
      self.loadstats()
      self.data['Restart-at']=self.lastsave
      self.data['Hostname']=socket.gethostname()
      self.masterpid=os.fork()
      if self.masterpid==0: # main process
        signal.signal(signal.SIGTERM,self.sigterm)
        self.pipe_r,self.pipe_w=os.pipe()
        debug.echo(1,"statistics collector started")
        while 1:
          try:
            self.conn,addr=self.s.accept()
            socket_settimeout(self.conn,10)
            signal.signal(signal.SIGUSR1,self.sighandler)
            signal.signal(signal.SIGCHLD,signal.SIG_IGN)
            pid=os.fork()
            if pid==0:
              try:
                f=self.conn.makefile("rw",0)
                req=f.readline()
                reg1=re.search("^GET /(.*) HTTP/1.[01]",req)
                if reg1:
                  dirs=reg1.group(1).split("/")
                  args=self.parseargs(dirs[1:])
                  lines={}
                  while 1:
                    line=f.readline()
                    if (line=="\r\n") or (line=="\n"):
                      break
                    a=line.split(": ")
                    if len(a)>1:
                      lines[a[0]]=a[1].strip()
                    else:
                      lines[a[0]]=a[0].strip()
                  if dirs[0]=="update":
                    try:
                      if lines["Authorization"]==ACCESSPASS:
                        d=string.join(dirs[1:],"/")+"\n"+' '*1024
                        os.write(self.pipe_w,d[0:1024])
                        self.message("200 OK")
                        self.conn.shutdown(2)
                        self.conn.close()
                        os.kill(os.getppid(),signal.SIGUSR1)
                        os._exit(2)
                      else:
                        self.message("401 Authorization Required")
                    except KeyError:
                      self.message("401 Authorization Required")
                  else:
                    self.conn.send("HTTP/1.1 200 OK\r\n")
                    self.conn.send("Content-Type: text/plain; charset=iso-8859-1\r\n")
                    self.conn.send("\r\n");
                    self.data["Uptime"]=uptime(self.data['Start-at'])
                    self.data["Restart-age"]=uptime(self.data['Restart-at'])
                    sstr=''
                    if dirs[0]=="":
                      for key,value in self.data.items():
                        sstr+=key+":\t"+str(value)+"\r\n"
                      self.conn.send(sstr)
                    elif dirs[0]=="mrtg":
                      for key in dirs[1:]:
                        if key in self.data.keys():
                          try:
                            o="%-20.0f" % self.data[key]
                            sstr+=o.strip()+"\r\n"
                          except TypeError:
                            sstr+=str(self.data[key])+"\r\n"
                        else: # return 0 for unknown key
                          sstr+="0\r\n"
                      self.conn.send(sstr)
                    else:
                      self.message("404 Not Found")
                else:
                  self.message("501 Method Not Implemented")
                  self.conn.send(req)
              except socket.error,(ec,es):
                if not ec in [32,104,107]:
                  debug.echo(3,"collector: socket.error: ",ec,es)
                  debug.traceback(4,"collector: ")
              except:
                debug.traceback(3,"collector: ")
              try:
                self.conn.shutdown(2)
                self.conn.close()
              except socket.error:
                pass 
              os._exit(0)
            else: # in parent
              self.childs.append(pid)
          except os.error,(ec,es):
            try:
              self.conn.shutdown(2)
              self.conn.close()
            except:
              pass
          except:
            pass
        # while 1 ... end
      atexit.register(self.killself)
  def sighandler(self,sn,stack):
      try:
        signal.alarm(2)
        str=os.read(self.pipe_r,1024)
        signal.alarm(0)
      except OSError,(ec,es):
        if ec==4: # Interrupted system call
          debug.echo(3,"collector(): ",es)
          return ''
      args=self.parseargs(str.split("/"))
      for k,v in args.items():
        try:
          self.data[k]+=float(v)
        except KeyError:
          self.data[k]=float(v)
        except ValueError:
          self.data[k]=v
        except TypeError:
          debug.echo(0,"collector(): TypeError: ",k,v,type(v))
      # save stats evry 5 minutes
      ctime=time.mktime(time.localtime())
      if (ctime-self.lastsave)>300:
        self.lastsave=ctime
        self.savestats()
  def sigterm(self,sn,stack):
      if self.masterpid==0:
        self.savestats()
      debug.echo(1,"collector(): Exiting - SIGTERM ...")
      os._exit(0)
  def loadstats(self):
      try:
        f=open('/tmp/status','r')
        for line in f.readlines():
          a=line.split(":\t")
          try:
            self.data[a[0]]=float(a[1])
          except ValueError:
            self.data[a[0]]=a[1]
        f.close()
      except:
        pass
  def savestats(self):
      f=open('/tmp/status','w')
      for k,v in self.data.items():
        f.write(k+":\t"+str(v)+"\n")
      f.close()
      debug.echo(4,"collector(): Status saved.")
  def killself(self):
      if os.getpid()==MASTER_PID:
        os.kill(self.masterpid,signal.SIGTERM)
  def message(self,str):
      self.conn.send("HTTP/1.1 "+str+"\r\n\r\n")
  def parseargs(self,dirs):
      args={}
      for arg in dirs:
        a=arg.split("=")
        if len(a)>1:
          args[urllib.unquote_plus(a[0])]=urllib.unquote_plus(a[1])
        else:
          args[urllib.unquote_plus(a[0])]='1'
      return args

def query(qt,args):
    if COLLECTOR_SERVER:
      s=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
      socket_settimeout(s,10)
      s.connect(COLLECTOR_SERVER)
      urlargv=''
      for arg in args:
        urlargv+="/"+urllib.quote_plus(arg,"/=")
      s.send("GET /"+qt+urlargv+" HTTP/1.1\r\n")
      s.send("Authorization: "+ACCESSPASS+"\r\n")
      s.send("\r\n")
      f=s.makefile("r",0)
      lines=[]
      while 1:
        line=f.readline()
        if (line=="\r\n") or (line=="\n"):
          break;
        lines.append(line)
      lines=[]
      str=''
      str+=s.recv(9999).rstrip()
      s.shutdown(2)
      return str
    else:
      return

def update(bs,t,infected=0,tempfail=0):
    '''update statistics by 'bs' bytes, 't' time,
       if infected, then update also filtered bytes,files
    '''
    args=['Total-count','Total-bytes='+str(bs),'Total-time='+str(t),
          'Temp-fail='+str(tempfail)]
    if infected:
      args.extend(['Filtered-count','Filtered-bytes='+str(bs)])
    try:
      query('update',args+status.collect)
      status.collect=[]
    except:
      pass

class status(base.match_any):
  '''
  status() interscanner
  
  Usage: status("String",scanner1(),scanner2(),...)
  
  Where: "String" is a string, which defines a prefix for status update
         scanner1(),... are sagator's scanners
  
  This interscanner can be used to collect some other statistics.
  Preffered usage:
    SCANNERS=[
      status("Virus",virus_scanner1(),...),
      status("Spam",spam_scanner1(),...)
    ]
  This collects Virus-count,Virus-bytes,Spam-count,Spam-bytes in collector.
  '''
  name='' # making this scanner invisible
  collect=[]
  def __init__(self,string,*scanners):
      self.status_string=string
      base.match_any.__init__(self,scanners)
  def scanstream(self,stream):
      vir,ret=base.match_any.scanstream(self,stream)
      if vir:
        status.collect.extend([self.status_string+"-count",
          self.status_string+"-bytes="+str(len(stream))])
      return vir,ret

if __name__ == '__main__':
  if len(sys.argv)>1:
    from config import COLLECTOR_SERVER
    print query(sys.argv[1],sys.argv[2:])
